@@ -1,11 +1,11 @@
 from functools import lru_cache
-from typing import Any, Callable, Dict, List, Optional, Tuple, Union
+from typing import Any, Callable, Optional, Union
 
 from .errors import InvalidSignatureError, SignatureBodyMismatchError
 from .validators import is_object_path_valid
 
 
-class SignatureType:
+class SignatureType:  # noqa: PLW1641
     """A class that represents a single complete type within a signature.
 
     This class is not meant to be constructed directly. Use the :class:`SignatureTree`
@@ -21,15 +21,25 @@ class SignatureType:
     """
 
     _tokens = "ybnqiuxtdsogavh({"
-    __slots__ = ("token", "children", "_signature")
+    __slots__ = (
+        "_child_0",
+        "_child_1",
+        "_signature",
+        "children",
+        "token",
+        "token_as_int",
+    )
 
     def __init__(self, token: str) -> None:
         """Init a new SignatureType."""
         self.token: str = token
-        self.children: List[SignatureType] = []
+        self.token_as_int = ord(token)
+        self.children: list[SignatureType] = []
+        self._child_0: Optional[SignatureType] = None
+        self._child_1: Optional[SignatureType] = None
         self._signature: Optional[str] = None
 
-    def __eq__(self, other: Any) -> bool:
+    def __eq__(self, other: object) -> bool:
         """Compare this type to another type or signature string."""
         if type(other) is SignatureType:
             return self.signature == other.signature
@@ -59,8 +69,20 @@ class SignatureType:
         self._signature = self._collapse()
         return self._signature
 
+    def _add_child(self, child: "SignatureType") -> None:
+        """Add a child type to this type.
+
+        :param child: The child type to add.
+        :type child: :class:`SignatureType`
+        """
+        if self._child_0 is None:
+            self._child_0 = child
+        elif self._child_1 is None:
+            self._child_1 = child
+        self.children.append(child)
+
     @staticmethod
-    def _parse_next(signature: str) -> Tuple["SignatureType", str]:
+    def _parse_next(signature: str) -> tuple["SignatureType", str]:
         if not signature:
             raise InvalidSignatureError("Cannot parse an empty signature")
 
@@ -75,16 +97,16 @@ class SignatureType:
             (child, signature) = SignatureType._parse_next(signature[1:])
             if not child:
                 raise InvalidSignatureError("missing type for array")
-            self.children.append(child)
+            self._add_child(child)
             return (self, signature)
-        elif token == "(":
+        if token == "(":
             self = SignatureType("(")
             signature = signature[1:]
             while True:
                 (child, signature) = SignatureType._parse_next(signature)
                 if not signature:
                     raise InvalidSignatureError('missing closing ")" for struct')
-                self.children.append(child)
+                self._add_child(child)
                 if signature[0] == ")":
                     return (self, signature[1:])
         elif token == "{":
@@ -93,13 +115,13 @@ class SignatureType:
             (key_child, signature) = SignatureType._parse_next(signature)
             if not key_child or len(key_child.children):
                 raise InvalidSignatureError("expected a simple type for dict entry key")
-            self.children.append(key_child)
+            self._add_child(key_child)
             (value_child, signature) = SignatureType._parse_next(signature)
             if not value_child:
                 raise InvalidSignatureError("expected a value for dict entry")
             if not signature or signature[0] != "}":
                 raise InvalidSignatureError('missing closing "}" for dict entry')
-            self.children.append(value_child)
+            self._add_child(value_child)
             return (self, signature[1:])
 
         # basic type
@@ -130,7 +152,7 @@ class SignatureType:
             raise SignatureBodyMismatchError(
                 f'DBus INT16 type "n" must be Python type "int", got {type(body)}'
             )
-        elif body > INT16_MAX or body < INT16_MIN:
+        if body > INT16_MAX or body < INT16_MIN:
             raise SignatureBodyMismatchError(
                 f'DBus INT16 type "n" must be between {INT16_MIN} and {INT16_MAX}'
             )
@@ -142,7 +164,7 @@ class SignatureType:
             raise SignatureBodyMismatchError(
                 f'DBus UINT16 type "q" must be Python type "int", got {type(body)}'
             )
-        elif body > UINT16_MAX or body < UINT16_MIN:
+        if body > UINT16_MAX or body < UINT16_MIN:
             raise SignatureBodyMismatchError(
                 f'DBus UINT16 type "q" must be between {UINT16_MIN} and {UINT16_MAX}'
             )
@@ -154,7 +176,7 @@ class SignatureType:
             raise SignatureBodyMismatchError(
                 f'DBus INT32 type "i" must be Python type "int", got {type(body)}'
             )
-        elif body > INT32_MAX or body < INT32_MIN:
+        if body > INT32_MAX or body < INT32_MIN:
             raise SignatureBodyMismatchError(
                 f'DBus INT32 type "i" must be between {INT32_MIN} and {INT32_MAX}'
             )
@@ -166,7 +188,7 @@ class SignatureType:
             raise SignatureBodyMismatchError(
                 f'DBus UINT32 type "u" must be Python type "int", got {type(body)}'
             )
-        elif body > UINT32_MAX or body < UINT32_MIN:
+        if body > UINT32_MAX or body < UINT32_MIN:
             raise SignatureBodyMismatchError(
                 f'DBus UINT32 type "u" must be between {UINT32_MIN} and {UINT32_MAX}'
             )
@@ -178,7 +200,7 @@ class SignatureType:
             raise SignatureBodyMismatchError(
                 f'DBus INT64 type "x" must be Python type "int", got {type(body)}'
             )
-        elif body > INT64_MAX or body < INT64_MIN:
+        if body > INT64_MAX or body < INT64_MIN:
             raise SignatureBodyMismatchError(
                 f'DBus INT64 type "x" must be between {INT64_MIN} and {INT64_MAX}'
             )
@@ -190,7 +212,7 @@ class SignatureType:
             raise SignatureBodyMismatchError(
                 f'DBus UINT64 type "t" must be Python type "int", got {type(body)}'
             )
-        elif body > UINT64_MAX or body < UINT64_MIN:
+        if body > UINT64_MAX or body < UINT64_MIN:
             raise SignatureBodyMismatchError(
                 f'DBus UINT64 type "t" must be between {UINT64_MIN} and {UINT64_MAX}'
             )
@@ -204,10 +226,10 @@ class SignatureType:
     def _verify_unix_fd(self, body: Any) -> None:
         try:
             self._verify_uint32(body)
-        except SignatureBodyMismatchError:
+        except SignatureBodyMismatchError as ex:
             raise SignatureBodyMismatchError(
                 'DBus UNIX_FD type "h" must be a valid UINT32'
-            )
+            ) from ex
 
     def _verify_object_path(self, body: Any) -> None:
         if not is_object_path_valid(body):
@@ -295,7 +317,7 @@ class SignatureType:
 
         return True
 
-    validators: Dict[str, Callable[["SignatureType", Any], None]] = {
+    validators: dict[str, Callable[["SignatureType", Any], None]] = {
         "y": _verify_byte,
         "b": _verify_boolean,
         "n": _verify_int16,
@@ -315,7 +337,7 @@ class SignatureType:
     }
 
 
-class SignatureTree:
+class SignatureTree:  # noqa: PLW1641
     """A class that represents a signature as a tree structure for conveniently
     working with DBus signatures.
 
@@ -327,16 +349,19 @@ class SignatureTree:
     :ivar ~.signature: The signature of this signature tree.
     :vartype ~.signature: str
 
+    :ivar root_type: The root type of this signature tree.
+    :vartype root_type: :class:`SignatureType
+
     :raises:
         :class:`InvalidSignatureError` if the given signature is not valid.
     """
 
-    __slots__ = ("signature", "types")
+    __slots__ = ("root_type", "signature", "types")
 
     def __init__(self, signature: str = "") -> None:
         self.signature = signature
 
-        self.types: List[SignatureType] = []
+        self.types: list[SignatureType] = []
 
         if len(signature) > 0xFF:
             raise InvalidSignatureError("A signature must be less than 256 characters")
@@ -345,12 +370,14 @@ class SignatureTree:
             (type_, signature) = SignatureType._parse_next(signature)
             self.types.append(type_)
 
-    def __eq__(self, other: Any) -> bool:
+        self.root_type = self.types[0] if self.types else None
+
+    def __eq__(self, other: object) -> bool:
         if type(other) is SignatureTree:
             return self.signature == other.signature
         return super().__eq__(other)
 
-    def verify(self, body: List[Any]) -> bool:
+    def verify(self, body: list[Any]) -> bool:
         """Verifies that the give body matches this signature tree
 
         :param body: the body to verify for this tree
@@ -375,7 +402,7 @@ class SignatureTree:
         return True
 
 
-class Variant:
+class Variant:  # noqa: PLW1641
     """A class to represent a DBus variant (type "v").
 
     This class is used in message bodies to represent variants. The user can
@@ -393,7 +420,7 @@ class Variant:
         :class:`SignatureBodyMismatchError` if the signature does not match the body.
     """
 
-    __slots__ = ("type", "signature", "value")
+    __slots__ = ("signature", "type", "value")
 
     def __init__(
         self,
@@ -402,14 +429,6 @@ class Variant:
         verify: bool = True,
     ) -> None:
         """Init a new Variant."""
-        self._init_variant(signature, value, verify)
-
-    def _init_variant(
-        self,
-        signature: Union[str, SignatureTree, SignatureType],
-        value: Any,
-        verify: bool,
-    ) -> None:
         if type(signature) is SignatureTree:
             signature_tree = signature
             self.signature = signature_tree.signature
@@ -434,15 +453,21 @@ class Variant:
                 )
             self.type.verify(value)
 
-    def __eq__(self, other: Any) -> bool:
+    @staticmethod
+    def _factory(signature_tree: SignatureTree, value: Any) -> "Variant":
+        self = Variant.__new__(Variant)
+        self.signature = signature_tree.signature
+        self.type = signature_tree.root_type
+        self.value = value
+        return self
+
+    def __eq__(self, other: object) -> bool:
         if type(other) is Variant:
             return self.signature == other.signature and self.value == other.value
         return super().__eq__(other)
 
     def __repr__(self) -> str:
-        return "<dbus_fast.signature.Variant ('{}', {})>".format(
-            self.type.signature, self.value
-        )
+        return f"<dbus_fast.signature.Variant ('{self.type.signature}', {self.value})>"
 
 
 get_signature_tree = lru_cache(maxsize=None)(SignatureTree)
